import itertools
import random
import time
from collections import defaultdict
from http import HTTPStatus
from os import environ

from requests.exceptions import RequestException, JSONDecodeError
from typing import Union

from exceptions import ExternalServiceException
from requests_adapter import requests_retry_session
from static.config import ENVIRONMENT, SINGLETON
from static.utils import is_test_env, Log

logger = Log()


class CSR:
    def __init__(self, env: str):
        """This class interacts with CBC based on the env passed in (dev01,test03, etc)"""

        self.org_keys = [
            "orgKey",
            "orgType",
            "mode",
            "trialEndTime",
            "paidStartTime",
            "createTime",
            "name",
            "id",
            "orgUuid",
            "active",
            "lastLoginTime",
            "lastUpdateTime",
        ]
        self.env = env.upper()
        self.singleton_overrides = SINGLETON.__members__.keys()

        self.adminservices = "{}://{}/adminServices/".format(
            self._get_ingress_protocol("adminServices"), self._get_ingress_host_name("adminServices")
        )
        self.appservices = "{}://{}/appservices/".format(
            self._get_ingress_protocol("appservices"), self._get_ingress_host_name("appservices")
        )
        self.retention = "{}://{}/provisioning/retention/".format(
            self._get_ingress_protocol("provisioning/retention"), self._get_ingress_host_name("provisioning/retention")
        )
        self.vulnerability = "{}://{}/vulnerability/".format(
            self._get_ingress_protocol("vulnerability"), self._get_ingress_host_name("vulnerability")
        )
        self.compliance = "{}://{}/compliance/".format(
            self._get_ingress_protocol("compliance"), self._get_ingress_host_name("compliance")
        )
        self.entitlements = "{}://{}/entitlements/".format(
            self._get_ingress_protocol("entitlements"), self._get_ingress_host_name("entitlements")
        )
        self.watchlists = "{}://{}/api/watchlistmgr/".format(
            self._get_ingress_protocol("api/watchlistmgr"), self._get_ingress_host_name("api/watchlistmgr")
        )
        self.identity = "{}://{}/identity/".format(
            self._get_ingress_protocol("identity"), self._get_ingress_host_name("identity")
        )
        self.policyservice = "{}://{}/policyservice/".format(
            self._get_ingress_protocol("policyservice"), self._get_ingress_host_name("policyservice")
        )
        self.rules = "{}://{}/rules/private/".format(
            self._get_ingress_protocol("rules/private"), self._get_ingress_host_name("rules/private")
        )

        self.requests = requests_retry_session()  # retry mechanism for requests

    def _get_ingress_host_name(self, endpoint) -> str:
        if self.env in self.singleton_overrides:
            if endpoint == "appservices":
                return "gateway.psc.internal:8900"

        return environ.get("INTERNAL_INGRESS_HOSTNAME")

    def _get_ingress_protocol(self, endpoint) -> str:
        if self.env in self.singleton_overrides:
            if endpoint == "appservices":
                return "http"

        return environ.get("INTERNAL_INGRESS_PROTOCOL") or ("http" if is_test_env() else "https")

    def _get_connector(self) -> str:
        api_key = environ.get("API_KEY")
        connector_id = environ.get("CONNECTOR_ID")
        if self.env in self.singleton_overrides:
            api_key = environ.get(f"{self.env}_API_KEY") or api_key
            connector_id = environ.get(f"{self.env}_CONNECTOR_ID") or connector_id

        return f"{api_key}/{connector_id}"

    def _get_headers(self, url):
        headers = {
            "Accept": "application/json, text/plain, */*",
            "Content-Type": "application/json;charset=UTF-8",
            "X-Auth-Token": self._get_connector(),
            "User-Agent": "integration-clm",
        }

        headers.update(self._get_headers_overrides_per_env())
        headers.update(self._get_headers_overrides_per_url(url))

        return headers

    def _get_headers_overrides_per_env(self) -> dict:
        overrides_per_env = {
            ENVIRONMENT.DEV01.value: {"Host": "defense-dev01.cbdtest.io"},
            ENVIRONMENT.TEST03.value: {"Host": "defense-test03.cbdtest.io"},
            ENVIRONMENT.PROD01.value: {"Host": "dashboard.confer.net"},
            ENVIRONMENT.PROD02.value: {"Host": "defense.conferdeploy.net"},
            ENVIRONMENT.PROD05.value: {"Host": "defense-prod05.conferdeploy.net"},
            ENVIRONMENT.EAP01.value: {"Host": "defense-eap01.conferdeploy.net"},
            ENVIRONMENT.PROD06.value: {"Host": "defense-eu.conferdeploy.net"},
            ENVIRONMENT.PRODSYD.value: {"Host": "defense-prodsyd.conferdeploy.net"},
            ENVIRONMENT.PRODNRT.value: {"Host": "defense-prodnrt.conferdeploy.net"},
        }

        if self.env in overrides_per_env.keys():
            return overrides_per_env[self.env]

        return {}

    def _get_headers_overrides_per_url(self, url) -> dict:
        api_key = environ.get("API_KEY")
        connector_id = environ.get("CONNECTOR_ID")
        # Entitlements/Retention/Compliance/vulnerability/rules/watchlists service is singleton in use1 region,
        # so we want to override with the prod05 headers
        singleton_service_overrides = [
            self.entitlements,
            self.retention,
            self.compliance,
            self.vulnerability,
            self.rules,
            self.policyservice,
            self.watchlists,
        ]
        override_config = {
            SINGLETON.PROD01.value: {"Host": "defense-prod05.conferdeploy.net", "X-Auth-Token": f"{api_key}/{connector_id}"},
            SINGLETON.PROD02.value: {"Host": "defense-prod05.conferdeploy.net", "X-Auth-Token": f"{api_key}/{connector_id}"},
            SINGLETON.EAP01.value: {"Host": "defense-prod05.conferdeploy.net", "X-Auth-Token": f"{api_key}/{connector_id}"},
        }

        if _ := [item for item in singleton_service_overrides if item in url]:
            if self.env in override_config.keys():
                return override_config[self.env]

        return {}

    def send_request(
        self, method: str, url: str, data: dict = None, parameters: dict = None, bypass_codes: list = []
    ) -> Union[dict, str, None]:
        """central function used to send all api calls to CBC"""

        try:
            response = self.requests.request(
                method, url, headers=self._get_headers(url), json=data, params=parameters, verify=True
            )

            try:
                return response.json()
            except JSONDecodeError:
                return response.text

        except RequestException as e:
            logger.send_exception(e)

            if e.response is not None and e.response.status_code in bypass_codes:
                return None

            raise ExternalServiceException(f"Call to underlying service failed: {str(e)}", request=e.request, response=e.response)

    def get_org(self, org_id: str, codes: list = [HTTPStatus.FORBIDDEN]) -> dict:
        result = {}
        if org := self.send_request("GET", f"{self.appservices}v5/orgs/{org_id}", bypass_codes=codes):
            if org.get("organization"):
                result = {k: v for k, v in org["organization"].items() if k in self.org_keys}
                result.update({k: v for k, v in org.items() if k in self.org_keys})

        return result

    def get_org_by_name(self, org_name: str) -> dict:
        result = {}

        query = self.send_request("GET", f"{self.adminservices}v5/orgs", parameters={"org_mode": "ALL", "q": org_name})
        if query.get("organizations"):
            if id := next(
                (item["id"] for item in query["organizations"] if org_name.casefold() == item["name"].casefold()), None
            ):
                result = self.get_org(id)

        return result

    def get_org_by_uuid(self, org_uuid: str, codes: list = [HTTPStatus.NOT_FOUND, HTTPStatus.FORBIDDEN]) -> dict:

        query = self.send_request("GET", f"{self.identity}v1/orgs/{org_uuid}", bypass_codes=codes)

        # kinda messy but easier to handle in middle functions
        return query if isinstance(query, dict) else {}

    def post_org(self, data: dict) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs", data)

    def delete_org(self, org_id: str) -> dict:
        return self.send_request("DELETE", f"{self.adminservices}v5/orgs/{org_id}")

    def post_org_users(self, org_key: str, data: dict) -> dict:
        return self.send_request("POST", f"{self.appservices}v6/orgs/{org_key}/users", data)

    def post_org_delegation(self, org_id: str, mssp_id: str) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/delegations/{mssp_id}")

    def get_global_feature_flags(self) -> dict:
        return self.send_request("GET", f"{self.adminservices}v5/orgs/settings/ui-feature-flags")

    def get_org_feature_flags(self, org_id: str) -> dict:
        return self.send_request("GET", f"{self.adminservices}v5/orgs/{org_id}/settings/ui-feature-flags")

    def get_all_feature_flags(self, org_id: str) -> dict:
        org = self.get_org_feature_flags(org_id)["uifeatureData"]
        default = self.get_global_feature_flags()["uifeatureData"]

        default.update(org)

        return {"uifeatureData": default}

    def verify_feature_flags(self, org_id: str, new_flags: dict, recurse: bool = False) -> bool:
        time.sleep(random.uniform(0.75, 1.25))
        all_flags = self.get_all_feature_flags(org_id)["uifeatureData"]

        if not new_flags.items() <= all_flags.items():
            logger.send_error(str({"CLM": new_flags, "CBC": all_flags}))

            if recurse:
                self.post_org_feature_flags(org_id, new_flags)

            return False

        return True

    def post_org_feature_flags(self, org_id: str, new_flags: dict) -> dict:
        existing_org_feature_flags = self.get_org_feature_flags(org_id)["uifeatureData"]
        merged_org_flags = {**existing_org_feature_flags, **new_flags}

        # This is specific to default flag behavior
        # Only applied when we are looking to change a feature flag back to global CBC value
        for k, v in dict(merged_org_flags).items():
            if v == "default":
                merged_org_flags.pop(k, None)

        resp = self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/settings/ui-feature-flags", merged_org_flags)

        if not bool(environ.get("VERIFY_CBC_FLAGS", 1)):
            self.verify_feature_flags(org_id, merged_org_flags, True)

        return resp

    def post_org_defense_config(self, org_id: str, action: bool) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/settings/disable-defense/{action}")

    def build_sensor_config(self, name: str, conditions: str, feature: str, boolean: bool, persist: bool = False):
        return {
            "name": name,
            "value": "1" if boolean else "0",
            "conditions": conditions,
            "description": feature,
            "persist": persist,
            "enabled": boolean,
        }

    def post_org_global_sensor_settings(self, org_id: str, action: str) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/settings/internal_sensor_settings/{action}")

    def post_org_sensor_settings(self, org_key: str, body: dict) -> dict:
        prop, index = self.get_org_sensor_settings(org_key), None

        if prop:
            for i, p in enumerate(prop["settings"]):
                # Only if config global is false set other config props to false
                if prop.get("enabled") is False:
                    p["enabled"] = False

                # Grab index of config prop being set
                if p["name"].casefold() == body["name"].casefold():
                    index = i

            if prop.get("enabled") is False and body.get("enabled") is True:
                prop["enabled"] = body["enabled"]

            # Remove config prop with same name
            if index is not None:
                prop["settings"].pop(index)

            prop["settings"].append(body)  # Populate new config prop

            self.send_request("POST", f"{self.rules}v1/sensor_settings/psc:org:{org_key}", prop)
        else:
            initial_prop = {"settings": [], "enabled": body["enabled"]}
            self.send_request("POST", f"{self.rules}v1/sensor_settings/psc:org:{org_key}", initial_prop)

        if not prop:  # This will repeat one time when initially configuring
            time.sleep(random.uniform(0.05, 0.10))
            self.post_org_sensor_settings(org_key, body)

    def get_org_sensor_settings(self, org_key: str, codes: list = [HTTPStatus.NOT_FOUND]) -> dict:
        if result := self.send_request("GET", f"{self.rules}v1/sensor_settings/psc:org:{org_key}", bypass_codes=codes):
            return result

        return None

    def get_org_watchlists(self, org_key: str) -> dict:
        _ = self.send_request("GET", f"{self.watchlists}v3/orgs/{org_key}/watchlists")
        return _["results"] if "results" in _ else []

    def delete_org_watchlist_tags(self, org_key: str) -> dict:
        return self.send_request("DELETE", f"{self.watchlists}v3/orgs/{org_key}/watchlists/tag")

    def delete_org_watchlist_alerts(self, org_key: str) -> dict:
        return self.send_request("DELETE", f"{self.watchlists}v3/orgs/{org_key}/watchlists/alert")

    def post_org_contract_status(self, org_id: str) -> dict:
        return self.send_request(
            "POST",
            f"{self.adminservices}v5/orgs/{org_id}/settings/contract-status/PAID/START",
        )

    def put_org_vulnerability_assessment(self, org_key: str, body: list) -> dict:
        return self.send_request("PUT", f"{self.vulnerability}assessment/api/v1/orgs/{org_key}/config", body)

    def put_org_compliance_assessment(self, org_key: str, body: list) -> dict:
        return self.send_request("PUT", f"{self.compliance}assessment/api/v1/orgs/{org_key}/config", body)

    def put_org_policyservice_features(self, org_key: str, body: list) -> dict:
        current = self.get_org_policyservice_features(org_key)

        dic = defaultdict(dict)
        if type(current) is dict:
            if "features" in current:
                for item in (current["features"], body):
                    for elem in item:
                        dic[elem["name"]].update(elem)

                current["features"] = list(dic.values())
                payload = current
            else:
                payload = {**current, **{"features": body}}
        else:
            payload = {"features": body}

        return self.send_request("PUT", f"{self.policyservice}private/v1/orgs/{org_key}/features", payload)

    def get_org_policyservice_features(self, org_key: str) -> dict:
        return self.send_request("GET", f"{self.policyservice}private/v1/orgs/{org_key}/features")

    def post_org_prevention_policy(self, org_id: str) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/policies/standard/ngav")

    def put_org_retention_class(self, org_key: str, body: dict) -> dict:
        return self.send_request("PUT", f"{self.retention}v1/orgs/{org_key}/classes/default", body)

    def post_org_prevention_downgrade(self, org_id: str) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/ngav/disable")

    def post_org_prevention_upgrade(self, org_id: str) -> dict:
        return self.send_request("POST", f"{self.adminservices}v5/orgs/{org_id}/ngav/upgrade")

    def delete_org_retention_class(self, org_key: str) -> dict:
        return self.send_request("DELETE", f"{self.retention}v1/orgs/{org_key}/classes/default")

    def get_features(self) -> dict:
        return self.send_request("GET", f"{self.entitlements}v1/features")

    def get_entitlement_feature_flags(self) -> list:
        features = self.get_features()

        return list(itertools.chain.from_iterable([f["config"]["flags"] for f in features if f["config"]["flags"]]))

    def get_org_features(self, org_key: str) -> dict:
        return self.send_request("GET", f"{self.entitlements}v1/orgs/{org_key}/features")

    def get_org_feature_details(self, org_key: str) -> dict:
        return self.send_request("GET", f"{self.entitlements}v1/orgs/{org_key}/details")

    def get_org_entitlements(self, org_key: str) -> list:
        entitlements = []

        url = f"{self.entitlements}v1/orgs/{org_key}/entitlements"
        resp = self.send_request("GET", url)

        while resp:
            entitlements.extend(resp)
            ent = resp[-1]
            ent_id = ent["entitlement_id"]
            resp = self.send_request("GET", f"{url}?next={ent_id}")

        return entitlements

    def post_org_entitlement(self, org_key: str, body: dict) -> dict:
        return self.send_request("POST", f"{self.entitlements}v1/orgs/{org_key}/entitlements", body)

    def put_org_entitlement(self, org_key: str, eid: str, body: dict) -> dict:
        return self.send_request("PUT", f"{self.entitlements}v1/orgs/{org_key}/entitlements/{eid}", body)

    def delete_org_entitlement(self, org_key: str, eid: str) -> dict:
        return self.send_request("DELETE", f"{self.entitlements}v1/orgs/{org_key}/entitlements/{eid}")

    def get_entitlement_catalog(self) -> dict:
        return self.send_request("GET", f"{self.entitlements}v1/catalog")

    def post_effective_entitlements(self, entitlements: list) -> list:
        return self.send_request("POST", f"{self.entitlements}v1/features/calculate", entitlements)

    def post_org_execution(self, org_key: str, body: dict) -> dict:
        return self.send_request("POST", f"{self.entitlements}v1/orgs/{org_key}/executions", body)

    def post_execution(self, body: dict) -> dict:
        return self.send_request("POST", f"{self.entitlements}v1/executions", body)

    def delete_org_entitlements(self, org_key: str) -> dict:
        return self.send_request("DELETE", f"{self.entitlements}v1/orgs/{org_key}/")

    def post_org_policyservice_enable_threathunter(self, org_key: str) -> dict:
        return self.send_request("POST", f"{self.policyservice}private/v1/orgs/{org_key}/features/eedr/_enable")

    def post_org_policyservice_disable_threathunter(self, org_key: str) -> dict:
        return self.send_request("POST", f"{self.policyservice}private/v1/orgs/{org_key}/features/eedr/_disable")
