from static.config import DELTA_SORT
from static.utils import Log
from models.entitlement import new_entitlement, new_merged_entitlement

from cbc import CSR

logger = Log()


def get_org_feature_records(event: dict) -> dict:
    """This function is a bit redundant, but it calls the csr class to return org feature records from entitle service"""

    csr = CSR(event["request_data"]["environment"])

    if orgKey := event.get("initial_state", {}).get("orgKey", None):
        return csr.get_org_features(orgKey)

    return []


def get_calculated_entitlements(event: dict, supported_operations: list = ["REGISTER", "ENTITLE", "REENTITLE"]) -> dict:
    """Take cumulative entitlements and submit them to entitlements to get the calculated features"""
    new_entitlements = []
    existing_internal_entitlements = []
    existing_external_entitlements = []
    deleted_internal_entitlements = []
    updated_internal_entitlements = []

    org_data = {
        "org_key": (
            event["ticket"]["initial_state"]["orgKey"]
            if event["ticket"]["request_data"]["operation"] != "REGISTER"
            else event["ticket"]["final_state"]["orgKey"]
        ),
        "external_urn": event["ticket"]["request_data"]["external_urn"],
        "source": event["ticket"]["request_data"]["source"],
        "origin": event["ticket"]["request_data"]["origin"],
        "custom": event["ticket"]["request_data"]["custom"],
        "operation": event["ticket"]["request_data"]["operation"],
        "csp_organization": (
            event["ticket"]["final_state"]["csp_organization"] if "csp_organization" in event["ticket"]["final_state"] else {}
        ),
    }

    csr = CSR(event["ticket"]["request_data"]["environment"])
    active_org_entitlements = csr.get_org_entitlements(org_data["org_key"])

    if org_data["operation"] in supported_operations:
        new_entitlements = [new_entitlement(org_data, ent).asdict() for ent in event["ticket"]["request_data"]["entitlements"]]

    [  # Finds entitlements based on the same origin and id
        (
            existing_internal_entitlements.append(e)
            if org_data["external_urn"].casefold() == e["external_id"].casefold()
            and org_data["origin"].casefold() == e["origin"].casefold()
            else existing_external_entitlements.append(e)
        )
        for e in active_org_entitlements
    ]

    created_internal_entitlements = [  # Newly created entitlements
        ent
        for ent in new_entitlements
        if not list(
            filter(lambda item: ent["product_id"].casefold() == item["product_id"].casefold(), existing_internal_entitlements)
        )
    ]

    [  # Updated and deleted entitlements
        (
            updated_internal_entitlements.append(new_merged_entitlement(old, new).asdict())
            if (new := next((e for e in new_entitlements if old["product_id"].upper() == e["product_id"].upper()), None))
            else deleted_internal_entitlements.append(old)
        )
        for old in existing_internal_entitlements
    ]

    result = {
        "NONE": existing_external_entitlements,
        "CREATE": created_internal_entitlements,
        "UPDATE": updated_internal_entitlements,
        "DELETE": deleted_internal_entitlements,
    }

    logger.send_info(str({"Calculated entitlements": result}))

    return result


def get_calculated_features(event: dict, calculated_entitlements: dict) -> list:
    """Take cumulative entitlements and submit them to entitlements to get the calculated features"""

    input = []
    [input.extend(i) for k, i in calculated_entitlements.items() if k != "DELETE"]

    csr = CSR(event["ticket"]["request_data"]["environment"])
    return csr.post_effective_entitlements(input)


def calculate_effective_feature_delta(event: dict) -> dict:
    """Calculate delta of features using the old features and the new ones"""

    initial_features = event.get("ticket").get("initial_state").get("features", []).copy()
    effective_features = event.get("ticket").get("final_state").get("calculated_features", []).copy()
    result = {}

    for feature in effective_features:
        item = next((f for f in initial_features if f["feature_id"] == feature["feature_id"]), None)

        if item == feature:
            feature["change"] = "NONE"
        elif item:
            feature["change"] = "UPDATE"
        else:
            feature["change"] = "CREATE"

        result[feature["feature_id"]] = feature

    # This must stay below effective features loop
    for feature in initial_features:
        item = next((cft for cft in effective_features if cft["feature_id"] == feature["feature_id"]), None)

        if not item:
            feature["change"] = "DELETE"
            result[feature["feature_id"]] = feature

    logger.send_info({"delta": result})
    return dict(sorted(result.items(), key=lambda tup: DELTA_SORT[tup[1]["change"]]))


def get_entitlement_delta(calculated_entitlements: dict, filter: str = "") -> dict:
    """Function that will update the entitlement delta format"""
    delta = []

    for key in calculated_entitlements.keys():
        if key != filter:
            for item in calculated_entitlements[key]:
                delta.append({"change": key, "record": item})

    return sorted(delta, key=lambda i: DELTA_SORT[i["change"]])


def update_entitlement_record(event: dict) -> dict:
    """Function that will update entitlements based on the delta calculated in CLM"""
    csr = CSR(event["ticket"]["request_data"]["environment"])

    for ent in event["ticket"]["final_state"]["entitlement_delta"]:
        if ent["change"] == "CREATE":
            csr.post_org_entitlement(ent["record"]["org_key"], ent["record"])
        elif ent["change"] == "UPDATE":
            csr.put_org_entitlement(ent["record"]["org_key"], ent["record"]["entitlement_id"], ent["record"])
        elif ent["change"] == "DELETE":
            csr.delete_org_entitlement(ent["record"]["org_key"], ent["record"]["entitlement_id"])

    return event["ticket"]


def update_entitlement_execution(event: dict) -> dict:
    csr = CSR(event["ticket"]["request_data"]["environment"])

    if "orgKey" in event["ticket"]["initial_state"]:
        event["ticket"]["final_state"]["features"] = csr.get_org_features(event["ticket"]["initial_state"]["orgKey"])
    elif "orgKey" in event["ticket"]["final_state"]:
        event["ticket"]["final_state"]["features"] = csr.get_org_features(event["ticket"]["final_state"]["orgKey"])

    execution = {
        "request_id": event["ticket"]["request_urn"],
        "origin": event["ticket"]["origin"],
        "system": event["ticket"]["source"],
        "operation": event["ticket"]["operation"],
        "status": event["ticket"]["execution_status"],
        "details": dict(event["ticket"]),
        "external_transaction_id": event["ticket"]["request_data"].get("external_transaction_id"),
    }

    if event["ticket"]["operation"] == "REGISTER":
        execution["org_name"] = event["ticket"]["request_data"]["org_name"]

        if "orgKey" in event["ticket"]["final_state"]:
            execution["org_key"] = event["ticket"]["final_state"]["orgKey"]
    else:
        execution["org_key"] = event["ticket"]["request_data"]["org_key"]

    logger.send_info(
        {"operation": event["ticket"]["operation"], "status": event["ticket"]["execution_status"], "value": execution}
    )

    if "org_key" in execution:
        csr.post_org_execution(execution["org_key"], execution)
    else:
        csr.post_execution(execution)

    return event["ticket"]
