from psc_orgkey import decode_orgkey
from static.utils import Log
from json import dumps
from os import environ
import requests
from exceptions import WebhookServiceException

logger = Log()


class Webhook:
    def __init__(
        self,
        url: str,
        request_urn: str,
        org_name: str = None,
        org_key: str = None,
        environment: str = None,
        error: dict = {},
    ):
        self.url = url
        self.request_urn = request_urn
        self.org_name = org_name
        self.org_key = org_key
        self.error = error
        self.region = environ.get("AWS_REGION")
        self.code = "Could not parse error, see details below."
        self.org_id = None
        self.org_environment = None

        try:
            self.org_id = decode_orgkey(orgkey=self.org_key).org_id
            self.org_environment = decode_orgkey(orgkey=self.org_key).env
        except Exception:
            logger.send_warning("Unable to decode org key for GChat Webhook.")

        try:
            self.code = self.error["Error"]
        except Exception:
            logger.send_warning("Unable pull error code from SM event.")

        if not self.org_environment:
            self.org_environment = environment

        self.link = f"https://{self.region}.console.aws.amazon.com/states/home?region={self.region}#/v2/executions/details/{self.request_urn}"

    def build(self):
        self.message = {
            "cardsV2": [
                {
                    "cardId": "unique-card-id",
                    "card": {
                        "header": {
                            "title": "integration-clm",
                            "subtitle": f"{self.request_urn.split(':')[-1]}",
                            "imageUrl": "https://developers.google.com/chat/images/quickstart-app-avatar.png",
                            "imageType": "CIRCLE",
                        },
                        "sections": [
                            {
                                "collapsible": False,
                                "widgets": [
                                    {
                                        "decoratedText": {
                                            "text": f"<b><font color='#FF2800'>FAILED</font></b><br><i>{self.code}</i>"
                                        },
                                    }
                                ],
                            },
                            {
                                "header": "Organization Details",
                                "collapsible": False,
                                "uncollapsibleWidgetsCount": 0,
                                "widgets": [
                                    {"decoratedText": {"topLabel": "Environment", "text": f"{self.org_environment}"}},
                                    {"decoratedText": {"topLabel": "Key", "text": f"{self.org_key}"}},
                                    {"decoratedText": {"topLabel": "Id", "text": f"{self.org_id}"}},
                                    {"decoratedText": {"topLabel": "Name", "text": f"{self.org_name}"}},
                                ],
                            },
                            {
                                "header": "Error Details",
                                "collapsible": True,
                                "uncollapsibleWidgetsCount": 0,
                                "widgets": [
                                    {
                                        "textParagraph": {"text": dumps(self.error, indent=2)},
                                    }
                                ],
                            },
                            {
                                "collapsible": False,
                                "widgets": [
                                    {
                                        "buttonList": {
                                            "buttons": [
                                                {
                                                    "text": "AWS Step Functions Link",
                                                    "color": {"red": 0, "green": 0, "blue": 0, "alpha": 0},
                                                    "onClick": {
                                                        "openLink": {"url": f"{self.link}"},
                                                    },
                                                }
                                            ]
                                        }
                                    }
                                ],
                            },
                        ],
                    },
                }
            ],
        }

    def fire(self):
        response = requests.post(url=self.url, headers={"Content-Type": "application/json; charset=UTF-8"}, json=self.message)

        if not response.status_code == 200:
            raise WebhookServiceException(url=response.url, status_code=response.status_code, body=response.json())

        return response
