import psc_orgkey
from static.utils import Log
from models.user import Users
from cbc import CSR

logger = Log()


def provision_admin_account(ticket: dict) -> list:
    """This will go through and update the super users for the corresponding org"""

    csr = CSR(ticket["request_data"]["environment"])
    result = []

    for contact in ticket.get("request_data").get("contacts", []):
        user = Users(**contact)
        user.assign_bcc_origin(ticket["request_data"]["origin"])  # Assign if bcc needed
        body = user.asdict()

        response = csr.post_org_users(ticket["final_state"]["orgKey"], body)

        # Response does not send back the role and email from what I recall
        response["role"] = body["role_urn"]
        response["email"] = body["email"]
        result.append(response)

    return result


def get_org(ticket: dict) -> dict:
    """this function uses the csr functions to find an org in the CBC"""

    csr = CSR(ticket["request_data"]["environment"])

    org_result = (
        csr.get_org_by_name(ticket["request_data"]["org_name"])
        if ticket["request_data"]["operation"] == "REGISTER"
        else csr.get_org(ticket["request_data"]["org_id"])
    )

    if org_result:
        org_result["stage"] = csr.get_org_feature_details(org_result.get("orgKey")).get("stage")

    return org_result


def create_organization(ticket: dict) -> dict:
    """this function uses the csr functions to create an org in the CBC"""

    csr = CSR(ticket["request_data"]["environment"])

    uuid = ticket.get("final_state", {}).get("csp_organization", {}).get("id")
    serial_number = ticket.get("request_data", {}).get("custom", {}).get("fast", {}).get("serialNumber")
    customer_name = ticket.get("request_data", {}).get("custom", {}).get("fast", {}).get("erpCustAcctName")
    org_name = f"{customer_name} ({serial_number})"

    org_create = csr.post_org(
        {
            "domain": ticket["request_data"]["org_name"],
            "orgUuid": uuid,
            "name": org_name if customer_name and serial_number else None,
        }
    )["org"]

    start_organization_contract(ticket, org_create["id"])
    assign_organization_mssp(ticket, org_create["id"])

    org_result = csr.get_org(org_create["id"], codes=[])
    if org_result:
        org_result["stage"] = csr.get_org_feature_details(org_result.get("orgKey")).get("stage")

    return org_result


def deregister_organization(ticket: dict) -> bool:
    """this function uses the csr functions to delete an org in CBC"""

    csr = CSR(ticket["request_data"]["environment"])
    org_entitlements_delete = csr.delete_org_entitlements(ticket["request_data"]["org_key"])
    logger.send_info(f"Org Delete Response for org {ticket['request_data']['org_key']}: {org_entitlements_delete}")
    org_delete = csr.delete_org(ticket["request_data"]["org_id"])
    logger.send_info(f"Org Delete Response for org {ticket['request_data']['org_id']}: {org_delete}")

    return True


def start_organization_contract(ticket: dict, org_id: str):
    """this function uses the csr functions to transition from TRIAL to PAID in CBC"""

    entitlement_evals = [bool(entitle["eval"]) for entitle in ticket["request_data"]["entitlements"] if "eval" in entitle]
    entitlement_full_trial = [
        bool(entitle.get("custom", {}).get("wlft", False)) for entitle in ticket["request_data"]["entitlements"]
    ]

    if ticket.get("initial_state").get("mode", None) == "PAID":
        logger.send_info("skipping contract because org is already set to PAID")
        return False

    if any(entitlement_full_trial):
        logger.send_info("wlft is set to true overriding contract")
    elif all(entitlement_evals):
        logger.send_info("skipping contract because entitlements only include TRIAL")
        return False

    csr = CSR(ticket["request_data"]["environment"])
    csr.post_org_contract_status(org_id)

    return True


def assign_organization_mssp(ticket: dict, org_id: str) -> bool:
    """this function uses the csr functions to assign a parent org in CBC"""

    # ! Only the source is able to update the MSSP
    # We will need to remove the MSSP if from originator and same id
    if parent_key := ticket["request_data"].get("parent", False):
        csr = CSR(ticket["request_data"]["environment"])
        csr.post_org_delegation(psc_orgkey.decode_orgkey(parent_key).org_id, org_id)

    return True


def get_flags(ticket: dict, org_only: False):

    csr = CSR(ticket["request_data"]["environment"])

    if org_only:
        flags = csr.get_org_feature_flags(ticket["initial_state"]["id"])["uifeatureData"]
    else:
        flags = csr.get_global_feature_flags()["uifeatureData"]

    config = csr.get_entitlement_feature_flags()

    return {k: v for k, v in flags.items() if k in config}
