from datetime import datetime
from uuid import UUID

import psc_orgkey
from jwt import decode as jwt_decode

from cbc import CSR
from exceptions import InternalServerError, ResourceAlreadyExists, ResourceNotFound, ValidationException
from static.utils import Log

logger = Log()


def validate_8601(date_str):
    date_str = str(date_str)
    try:
        return datetime.fromisoformat(date_str.replace("Z", "+00:00"))
    except ValueError:
        raise ValidationException(f"Invalid date must be in 8601 format: '{date_str}'")


def validate_uuid_format(uuid: str, version: int = 4):
    try:
        UUID(uuid, version=version)
    except ValueError:
        raise ValidationException("CSP uuid id must be a valid uuid4 format.")

    return True


def format_8601(date_str):
    date = datetime.fromisoformat(date_str.replace("Z", "+00:00"))
    return date.isoformat(sep="T", timespec="microseconds").split("+")[0] + "Z"


def validate_organization(event: dict):
    csp_bool = event.get("request_data").get("custom", {}).get("is_csp", None)
    register_csp_uuid = event.get("request_data").get("custom", {}).get("uuid")
    purge_csp_uuid = event.get("request_data", {}).get("org_uuid")
    org_key = event.get("request_data", {}).get("org_key", None)
    user_token = event.get("request_data", {}).get("user_token", None)
    origin = event.get("origin", "")
    operation = event.get("operation", "")
    csr = CSR(event["request_data"]["environment"])

    # CSP Checks
    if csp_bool is None:  # Check is CSP exists
        raise ValidationException("Custom CSP flag is required.")
    elif csp_bool:  # If CSP exists
        if not event["request_data"]["custom"].get("service_instance_id", None):  # if CSP service instance not there
            raise InternalServerError("Unable to determine service instance for environment.")

        if origin == "LOCAL" and user_token and operation == "REGISTER":
            try:
                jwt_decode(user_token, options={"verify_signature": False})["acct"]
            except Exception as e:
                logger.send_exception(f"Error encountered while decoding jwt claims: {str(e)}")
                raise ValidationException("CSP enabled organization can only be created from valid CSP account.")

    # Operation Checks
    if operation.casefold() == "register":
        if event["request_data"]["org_name"].count(".") < 1:
            raise ValidationException("Org name must be valid domain format.")

        if origin.casefold() == "vcdr" and register_csp_uuid:
            if register_csp_uuid:
                validate_uuid_format(register_csp_uuid)

    elif operation.casefold() == "purge":
        if origin.casefold() != "vcdr":
            raise ValidationException("Purge is not supported for your origin.")

        if purge_csp_uuid:
            validate_uuid_format(purge_csp_uuid)

            if query := csr.get_org_by_uuid(purge_csp_uuid):
                if query.get("org_key") != org_key:
                    raise ValidationException("Org key provided is not attached to the CSP organization, please check org key.")
            else:
                raise ResourceNotFound(
                    f"Organization cannot be purged because it does not exist in {event['request_data']['environment']}."
                )

            if entitlements := csr.get_org_entitlements(org_key):
                if origin.casefold() not in [ent["origin"].casefold() for ent in entitlements]:
                    raise ValidationException("You are unauthorized to perform a delete on this org.")

    return event


def validate_operation(event: dict):
    """this function will validate which type of request type it is and
    perform logic based on whether the org query return data or not
    """
    csr = CSR(event["ticket"]["request_data"]["environment"])

    request = event.get("ticket").get("request_data")
    initial = event.get("ticket").get("initial_state")
    register_csp_uuid = request.get("custom", {}).get("uuid")

    if request.get("operation") == "REGISTER":
        if initial.get("orgKey", None):
            raise ResourceAlreadyExists("Organization already exists.")

        if request.get("origin") == "VCDR" and register_csp_uuid:
            if query := csr.get_org_by_uuid(register_csp_uuid):
                if query.get("org_uuid"):
                    raise ResourceAlreadyExists("Organization already exists for uuid.")

    elif not request.get("operation") == "REGISTER":
        if not initial.get("orgKey", None):
            raise ResourceNotFound("Organization does not exist.")

    # This has to stay below the validate_organization because we can't save an execution without an org name
    if request.get("parent", None):
        try:
            return psc_orgkey.decode_orgkey(orgkey=request["parent"])
        except (psc_orgkey.InvalidOrgKeyError, ValueError):
            raise ValidationException("Invalid parent key supplied in request body.")

    if "active" in initial:  # Temp until across all envs
        if initial["active"] is False:
            raise ValidationException("Organization state must to be set to active.")

    return


def validate_added_entitlements(event: dict):
    csr = CSR(event["request_data"]["environment"])
    catalog = csr.get_entitlement_catalog()

    new_internal_entitlements = []
    current_internal_entitlements = []
    current_external_entitlements = []

    if org_key := event.get("request_data").get("org_key", None):
        current_entitlements = csr.get_org_entitlements(org_key)
        current_external_entitlements = [
            i
            for i in current_entitlements
            if i["origin"] != event["request_data"]["origin"] or i["external_id"] != event["request_data"]["external_urn"]
        ]
        current_internal_entitlements = [
            i
            for i in current_entitlements
            if i["external_id"] == event["request_data"]["external_urn"] and i["origin"] == event["request_data"]["origin"]
        ]

    for entitlement in event["request_data"]["entitlements"]:
        found = [
            i
            for i in catalog
            if entitlement["product"] == i["product_id"] and event["request_data"]["source"].casefold() == i["system"].casefold()
        ]

        if not found:
            raise ResourceNotFound(f"The product {entitlement['product']} does not exist in the catalog.")

        # custom attr validation fix
        for item in found[0]["features"]:
            if "retention_days" in item["config"]["custom_attrs"]:
                allowed_retention_days = item["config"]["custom_attrs"]["retention_days"]["options"]["default"]

                if (
                    "retention_days" not in entitlement["config"]
                    or entitlement["config"]["retention_days"] not in allowed_retention_days
                ):
                    raise ValidationException(f"Entitlement retention days must be one of {str(allowed_retention_days)}.")

        # there was a verbal request to remove this validation for inactive entitlements because of parity
        if event["request_data"]["operation"].casefold() == "entitle" and entitlement["status"].casefold() == "inactive":
            logger.send_info("Skipping validation for entitle + inactive status")
        else:
            # Needed for singleton entitlements
            if "end_date" in entitlement and "begin_date" in entitlement:
                if validate_8601(entitlement.get("end_date")) <= validate_8601(entitlement.get("begin_date")):
                    raise ValidationException("Entitlement end date must be greater or equal to begin date.")

        if not entitlement["source"] == event["request_data"]["source"]:
            raise ValidationException("Entitlement source must be the same.")

        if not entitlement["external_urn"] == event["request_data"]["external_urn"]:
            raise ValidationException("Entitlement external urn must be the same.")

        if event["request_data"]["operation"].casefold() == "register" and entitlement["status"].casefold() == "inactive":
            raise ValidationException("Entitlement status cannot be inactive for register operation.")

        if entitlement["status"].casefold() == "active":  # Products to be added
            new_internal_entitlements.append(entitlement)
        else:  # Products to be updated or removed
            if not [
                item
                for item in current_internal_entitlements
                if item["product_id"].casefold() == entitlement["product"].casefold()
            ]:
                raise ValidationException("Cannot update an entitlment that doesn't exist.")

    return {
        "current_internal_entitlements": new_internal_entitlements,
        "catalog": catalog,
        "current_external_entitlements": current_external_entitlements,
    }


def validate_entitlement_dependencies(data: dict):
    effective_features = []

    def extract_product_features(entitlements: list):
        product_features = []

        for entitlement in entitlements:
            if "product" in entitlement:
                product = entitlement.get("product")
            elif "product_id" in entitlement:
                product = entitlement.get("product_id")

            if "system" in entitlement:
                source = entitlement.get("system")
            elif "source" in entitlement:
                source = entitlement.get("source")

            for pdt in data["catalog"]:
                if product == pdt["product_id"] and source == pdt["system"] and entitlement["status"].casefold() == "active":
                    product_features.extend(pdt["features"])

        return product_features

    effective_features.extend(extract_product_features(data["current_internal_entitlements"]))
    effective_features.extend(extract_product_features(data["current_external_entitlements"]))
    feature_ids = [i["feature_id"] for i in effective_features]

    for item in effective_features:
        if item["conflicts"]:
            if item["conflicts_cmp_type"] == "ANY":
                if any(map(lambda v: v in feature_ids, item["conflicts"])):
                    raise ValidationException(
                        f"{item['feature_id']} conflicts with {item['conflicts_cmp_type']} of [{', '.join(item['conflicts'])}]."
                    )

            elif item["conflicts_cmp_type"] == "ALL":
                if all(map(lambda v: v in feature_ids, item["conflicts"])):
                    raise ValidationException(
                        f"{item['feature_id']} conflicts with {item['conflicts_cmp_type']} of [{', '.join(item['conflicts'])}]."
                    )

        if item["requires"]:
            if item["requires_cmp_type"] == "ANY":
                if not any(map(lambda v: v in feature_ids, item["requires"])):
                    raise ValidationException(
                        f"{item['feature_id']} requires with {item['requires_cmp_type']} of [{', '.join(item['requires'])}]."
                    )

            elif item["requires_cmp_type"] == "ALL":
                if not all(map(lambda v: v in feature_ids, item["requires"])):
                    raise ValidationException(
                        f"{item['feature_id']} requires with {item['requires_cmp_type']} of [{', '.join(item['requires'])}]."
                    )

    return


def validate_existing_entitlements(event: dict):
    csr = CSR(event["request_data"]["environment"])
    org_entitlements = csr.get_org_entitlements(event["request_data"]["org_key"])

    if len(event["request_data"]["entitlements"]) == 0:
        raise ValidationException("Global Reentitlement not available at this time.")

    if len(event["request_data"]["entitlements"]) > 1:
        raise ValidationException("A granular Reentitlement can only be performed on one product at a time.")

    for item in event["request_data"]["entitlements"]:
        available_entitlements = next(
            (e for e in org_entitlements if item["product"].casefold() == e["product_id"].casefold() and e["status"] == "ACTIVE"),
            None,
        )

        if not available_entitlements:
            raise ValidationException("Reentitle can only be performed on an active entitlement.")
    return
